#!/usr/bin/python3
import argparse
import os.path
import re
from mysqltotsv import Splitter, ExtractSchema, row_strip_quotes

def valid_file(inputfile):
    if not os.path.isfile(inputfile):
        raise argparse.ArgumentTypeError('mysql dump file does not exist')
    return inputfile

def valid_dir(outputdir):
    if not os.path.isdir(outputdir):
        raise argparse.ArgumentTypeError('output directory does not exist')
    return outputdir

arg_parser = argparse.ArgumentParser(description='Tool for conversion of large MySQL dumps to TSV format')
arg_parser.add_argument('--file', dest='file', action='store', required=True, type=valid_file, help='mysql dump file')
arg_parser.add_argument('--outdir', dest='outdir', action='store', required=True, type=valid_dir, help='output directory')
arg_parser.add_argument('--table-filter', dest='table_filter', action='store', required=False, type=str, help='filtered tables')
arg_parser.add_argument('--only-schema', dest='only_schema', action='store_true', help='write the schema to the output directory')
arg_parser.add_argument('--strip-quotes', dest='strip_quotes', action='store_true', help='strip quotes from values')
arg_parser.add_argument('--debug', dest='debug', action='store_true', help='print debug information')
arg_parser.add_argument('--ignore-errors', dest='ignore_errors', action='store_true', help='ignore processing errors')

args   = arg_parser.parse_args()

if args.only_schema:
    extract = ExtractSchema(args)
    extract.doit()
else:
    splitter = Splitter(args)
    rows_written = 0
    seen_outfile = set()

    batch_gen = splitter.next_batch()
    while True:
        batch = None
        try:
            batch = next(batch_gen)
        except StopIteration as e:
            break
        except Exception as e:
            if args.ignore_errors:
                print(e)
            else:
                raise e

        if args.ignore_errors and batch["parse_exc"]:
            continue

        outfile = os.path.join(args.outdir,batch["table_name"] + ".tsv")
        if os.path.exists(outfile) and outfile not in seen_outfile:
            os.remove(outfile)

        seen_outfile |= set([outfile])

        with open(outfile,"a") as fh_out:
            rows_written += len(batch["rows"])
            for r in batch["rows"]:
                if args.strip_quotes:
                    row_strip_quotes(r)
                fh_out.write("\t".join(map(str, r)) + "\n")
            print("rows written: ",rows_written)




import re
import os.path
import logging
import sys
from lark import Lark, logger as lark_logger
from lark import Transformer
sys.setrecursionlimit(50000)

parser = Lark(r"""
_STRING_INNER: /(.*?)/
_STRING_ESC_INNER: _STRING_INNER /(?<!\\)(\\\\)*?/
ESCAPED_STRING1: "'" _STRING_ESC_INNER "'"
ESCAPED_STRING2: "`" _STRING_ESC_INNER "`"

start: records+ ";" ?

records: record | record "," records

record: "(" cells ")"

cells: cell | cell "," cells

NULL: "NULL"

cell: NUMBER | ESCAPED_STRING1 | ESCAPED_STRING2 | NULL

%import common (WS_INLINE, NUMBER)
%ignore WS_INLINE
""", parser='lalr', start='start', regex=True)

class ProcessAST(Transformer):
    def records(self, records):
        if len(records) == 1:
            rv = tuple([records[0]])
        else:
            rv = records[0], *records[1]
        return rv

    def record(self, cells):
        return list(cells[0])

    def start(self, items):
        return items[0]

    def cells(self, cells):
        if len(cells) == 1:
            rv = cells[0]
        else:
            if isinstance(cells[1],(str,int,float,)):
                rv = cells[0],cells[1]
            elif isinstance(cells[1],tuple):
                rv = cells[0],*cells[1]
        return rv

    def cell(self, cell):
        if   re.match(r'^\d+(?:\.\d+)$', cell[0].value):
            return float(cell[0].value)
        elif re.match(r'^\d+$', cell[0].value):
            return int(cell[0].value)
        else:
            return re.sub(r"\t","",cell[0].value)

class ExtractSchema:
    def __init__(self, args):
        self.args = args

    def doit(self):
        self.in_fh = open(self.args.file,"rb")
        self.out_fh = open(os.path.join(self.args.outdir,"schema.sql"),"w")

        STATE1,STATE2=1,2
        state = 1

        buf = ""

        for line in self.in_fh:
            try:
                line = line.decode("utf-8")
            except UnicodeDecodeError as e:
                decoded_unicode = line.decode('utf-8', 'replace')
                line = decoded_unicode

            tbl_name = None
            g1 = re.match(r"^CREATE TABLE ", line)
            g2 = re.match(r"^\).*;", line)

            if state == STATE1 and g1:
                buf += line
                state = STATE2
            elif state == STATE2 and g2:
                buf += line
                state = STATE1
            elif state == STATE2:
                buf += line

        self.out_fh.write(buf)
        self.out_fh.close()
        self.in_fh.close()

class Splitter:
    def __init__(self, args):
        lark_logger.setLevel(logging.DEBUG)
        self.args = args
        if self.args.table_filter:
            self.table_filter = self.args.table_filter.split(",")
        else:
            self.table_filter = []

        self.fh = open(self.args.file,"rb")
        self.seen_table = set()

    def next_batch(self):
        num = 0
        for line in self.fh:
            try:
                line = line.decode("utf-8")
            except UnicodeDecodeError as e:
                decoded_unicode = line.decode('utf-8', 'replace')
                line = decoded_unicode

            tbl_name = None
            g = re.match(r"^INSERT INTO `([^\`]+)`", line)

            if g:
                tbl_name = g.groups()[0]

                if len(self.table_filter) > 0 and tbl_name not in self.table_filter:
                    continue
                num+=1
                if self.args.debug:
                    print("line num: ",num)

                line = re.sub(r"INSERT INTO `([^\`]+)` ","",line)
                line = re.sub(r" VALUES ",", ",line)
                line = re.sub(r";\s*$","", line)

                parse_exc = None
                rows = []
                try:
                    tree = parser.parse(line)
                    rows = list(ProcessAST().transform(tree))
                    if tbl_name in self.seen_table:
                        rows.pop(0)
                    self.seen_table |= set([tbl_name])
                except Exception as e:
                    parse_exc = e

                yield { "table_name": tbl_name, "rows": rows, "parse_exc": parse_exc }

        self.fh.close()



def row_strip_quotes(r):
    for i in range(len(r)):
        if isinstance(r[i], str):
            if (r[i].startswith("'")  and r[i].endswith("'")):
                r[i] = re.sub(r"^'","",re.sub(r"'$","",r[i]))
                continue
            if (r[i].startswith("\"") and r[i].endswith("\"")):
                r[i] = re.sub(r'^"',"",re.sub(r'"$',"",r[i]))
                continue
            if (r[i].startswith("`")  and r[i].endswith("`")):
                r[i] = re.sub(r'^`',"",re.sub(r'`$',"",r[i]))
                continue

